#!/usr/bin/env python
# Copyright 2013, 2104 Netflix, Inc.


import argparse
import copy
import functools
import imp
import importlib
import json
import logging
import os
import os.path
import pkgutil
import sys
import time
from contextlib import contextmanager

try:
    import gevent.monkey as gevent_monkey
except ImportError:
    gevent_monkey = None

NO_HELP = "No help :("

FUZZY_PARSED = []
RECENT_SUB_PARSERS = []
UNAVAILABLE_MODULES = []


class CustomDescriptionFormatter(argparse.RawTextHelpFormatter):
    def _get_help_string(self, action):
        help = action.help
        if "(default" not in help and type(action) not in (
            argparse._StoreConstAction,
            argparse._StoreTrueAction,
            argparse._StoreFalseAction,
        ):
            if action.default is not argparse.SUPPRESS:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    help += " (default: %(default)s)"
        return help


class ParserError(Exception):
    def __init__(self, parser, message):
        self.parser = parser
        self.message = message

    def report(self, force_message=None):
        if RECENT_SUB_PARSERS:
            RECENT_SUB_PARSERS[-1].print_help(sys.stderr)
        else:
            self.parser.print_help(sys.stderr)
        sys.stderr.write("\n")
        if force_message:
            sys.stderr.write(force_message)
        else:
            sys.stderr.write("%s: error: %s\n" % (self.parser.prog, self.message))
        sys.exit(2)


def split_args(args):
    """Split command line args in 3 groups:
    - head, containing the initial options
    - body, containing everything after head, up to the first option
    - tail, containing everything after body
    """

    head = []
    body = []
    tail = []
    for arg in args:
        if arg.startswith("-"):
            if body:
                tail.append(arg)
            else:
                head.append(arg)
        else:
            if tail:
                tail.append(arg)
            else:
                body.append(arg)
    return head, body, tail


def attempt_fuzzy_matching(args, candidates):
    head, body, tail = split_args(args)

    if not body:
        return None, None

    import re

    matches = []

    while body:
        partial_body = body
        fuzzy_body = ".*".join(partial_body)
        for candidate in candidates:
            if re.search(fuzzy_body, candidate):
                matches.append(candidate)
        if matches:
            break
        tail.insert(0, body.pop())

    if not matches:
        logging.debug("No fuzzy matches for command line args %s", args)
        return None, matches

    if len(matches) > 1:
        logging.debug("Multiple fuzzy matches for command line args %s: %s", args, matches)
        return None, matches

    new_args = []
    new_args.extend(head)
    new_args.extend(matches[0].split(" "))
    new_args.extend(tail)
    return new_args, matches


class BaseParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        argparse.ArgumentParser.__init__(self, *args, **kwargs)

    def error(self, message):
        raise ParserError(self, message)

    def parse_known_args(self, args=None, namespace=None):
        RECENT_SUB_PARSERS.append(self)
        return argparse.ArgumentParser.parse_known_args(self, args, namespace)

    def _check_value(self, action, value):
        if (
            action.choices is not None and value not in action.choices
        ):  # special handling for a nicer "choice" error message
            available = ", ".join(map(str, action.choices))
            if len(available) > 20:
                available = "\n%s" % available
            msg = "Invalid choice [%s], choose from: %s" % (value, available)
            raise argparse.ArgumentError(action, msg)
        else:
            super(BaseParser, self)._check_value(action, value)


class SmartCommandMapParser(BaseParser):
    def __init__(self, *args, **kwargs):
        super(SmartCommandMapParser, self).__init__(*args, **kwargs)
        self.root_sub = self.add_subparsers(help="Available sub-commands", parser_class=BaseParser, dest='sub-command')
        self.root_sub.required = True
        self.sub_map = {"": self.root_sub}
        self.flat_map = {}

    def add_namespace(self, namespace, parent):
        desc = "%s sub-command group" % namespace.capitalize()
        temp = parent.add_parser(
            namespace, help=desc, description=desc, formatter_class=CustomDescriptionFormatter, add_help=False
        )
        temp.add_argument("-h", "--help", dest="_help", action=argparse._HelpAction)
        sub = temp.add_subparsers(help="Available sub-commands", parser_class=BaseParser, dest='sub-command')
        sub.required = True
        self.sub_map[namespace] = sub
        return sub

    def add_item(self, subparser, module_name, item, command_path):
        name, node = item
        if node.get("type") == "cmd":
            parser = subparser.add_parser(
                name,
                help=node.get("help"),
                description=node.get("desc", node.get("help")),
                formatter_class=CustomDescriptionFormatter,
                add_help=False,
            )
            if node.get("error"):

                def _func(*args, **kwargs):
                    logging.error("This command is unavailable: %s", node.get("desc"))
                    logging.error("NB: after fixing the issue, remember to run oc refresh again")
                    sys.exit(1)

                parser.set_defaults(_func=_func)
            else:
                parser.set_defaults(_func=functools.partial(finish_parser, copy.copy(parser), module_name + "." + name))
            self.flat_map[(command_path + " " + name).strip()] = module_name + "." + name
        else:
            sub = self.add_namespace(name, subparser)
            for sub_node in node.items():
                self.add_item(sub, module_name + "." + name, sub_node, command_path + " " + name)

    def add_command_map(self, namespace, command_map):
        sub = self.sub_map.get(namespace, None)
        if sub is None:
            sub = self.add_namespace(namespace, self.root_sub)

        module_name = command_map["module"]
        for item in command_map["commands"].items():
            self.add_item(sub, module_name, item, namespace)

    def pre_parse_args(self, args):
        """Perform our first pass parsing of the given command line arguments.
        Try fuzzy matching if we can't parse the command line as is.
        Return the final args (eg. possibly corrected after fuzzy matching) and the actual command function to be executed.
        """
        try:
            parsed_args, _ = self.parse_known_args(args)
        except ParserError as pe:
            # if we get 'too few arguments here', user is referencing a command group, and we shouldn't fuzzy match
            if pe.message == "too few arguments":
                pe.report()
            logging.debug("Could not parse command line args [%s], attempting fuzzy matching", args)
            fixed_args, matches = attempt_fuzzy_matching(args, list(self.flat_map.keys()))
            if fixed_args:
                logging.info('Your input "%s" matches "%s"', " ".join(args), " ".join(fixed_args))
                FUZZY_PARSED.append((" ".join(args), " ".join(fixed_args)))
                args = fixed_args
                parsed_args, _ = self.parse_known_args(args)
            elif matches:
                message = "Your command line matched the following existing commands:\n    %s\n" % (
                    "\n    ".join(matches)
                )
                pe.report(force_message=message)
            else:
                pe.report()

        try:
            return args, parsed_args._func()
        except AttributeError:
            logging.error("known cligraphy bug with python 3, see https://github.com/cligraphy/cligraphy/issues/1")
            raise

    def parse_args(self, args=None):
        if args is None:
            args = sys.argv[1:]
        fixed_args, _ = self.pre_parse_args(args)
        return super(SmartCommandMapParser, self).parse_args(fixed_args)


def detect_monkey_patch():
    return gevent_monkey and len(gevent_monkey.saved) > 0


def error_module(module_name, exc):
    message = "%s: %s" % (exc.__class__.__name__, exc)
    UNAVAILABLE_MODULES.append((module_name, message))
    mod = imp.new_module(module_name)
    mod.__file__ = "__%s_error_stub__" % module_name
    mod.__doc__ = message
    mod.__error__ = True
    return mod


@contextmanager
def import_time_reporting(module_name):
    start = time.time()
    yield
    elapsed = time.time() - start
    if elapsed > 0.5:
        logging.info("slow import: module %s took %.2f seconds" % (module_name, elapsed))


def find_command_modules(prefix, path):
    """Returns a list of all command modules and packages"""
    prefix = "%s." % (prefix)
    for loader, module_name, is_pkg in pkgutil.iter_modules(path, prefix=prefix):
        try:
            with import_time_reporting(module_name):
                mod = importlib.import_module(module_name)

            if detect_monkey_patch():
                logging.error("gevent monkey patching detected after module %s was loaded", module_name)
                raise Exception("monkey patching is not allowed in cligraphy command modules")

            if getattr(mod, "main", None):
                yield mod

            if is_pkg:
                with import_time_reporting(module_name):
                    pkg = importlib.import_module(module_name)
                for mod in find_command_modules(pkg.__name__, pkg.__path__):
                    yield mod
        except KeyboardInterrupt:
            raise
        except ImportError as ie:
            yield error_module(module_name, ie)
        except SyntaxError as se:
            yield error_module(module_name, se)
        except BaseException as be:
            yield error_module(module_name, be)


def finish_parser(parser, module_name):
    logging.debug("Build actual parser for module %s", module_name)
    module = importlib.import_module(module_name)

    if hasattr(module, "configure"):
        module.configure(parser)

    parser.set_defaults(_func=module.main)
    parser.add_argument("-h", "--help", dest="_help", action=argparse._HelpAction)
    return module.main


class AutoDiscoveryCommandMap(object):
    """Automatically builds a commands map for our oc sub commands"""

    def __init__(self, cligraph, root_module_name):
        self.cligraph = cligraph
        self.root_node = {}
        self.package_nodes = {"": self.root_node}
        self.root_module_name = root_module_name

    def parse_help(self, module):
        """Parse docstring to generate usage"""
        doc = getattr(module, "__doc__", None)
        if doc is None:
            return (
                NO_HELP,
                "No description provided, you could add one! Code is probably located here: %s"
                % (module.__file__.replace("pyc", "py")),
            )
        else:
            halp, _, desc = doc.strip().replace("%", "%%").partition("\n")
            halp = halp.strip()
            if not desc:
                desc = halp
            return halp, desc

    def get_node(self, package_name):
        """Get or create a sub node"""
        logging.debug("Looking for package node [%s]", package_name)
        complete_package_name = package_name
        sub = self.package_nodes.get(package_name, None)
        if sub is None:
            logging.debug("-- not found")
            if "." in package_name:
                parent_name, _, package_name = package_name.partition(".")
                parent = self.get_node(parent_name)
            else:
                parent = self.root_node

            sub = parent[package_name] = self.package_nodes[complete_package_name] = {}
        return sub

    def build(self, force_autodiscover=False):
        root_module = importlib.import_module(self.root_module_name)

        cached_command_map_filename = os.getenv(
            "%s_COMMANDS_CACHE" % (self.cligraph.conf.tool.shortname.upper()),
            os.path.join(self.cligraph.conf.user.dotdir, "commands.json"),
        )
        if not force_autodiscover and os.path.exists(cached_command_map_filename):
            try:
                with open(cached_command_map_filename) as fpin:
                    command_map = json.load(fpin)
                return command_map
            except ValueError:
                logging.warning("Could not parse existing commands cache %s, ignoring it", cached_command_map_filename)

        for module in find_command_modules(self.root_module_name, root_module.__path__):
            package_name, _, module_name = module.__name__.rpartition(".")
            assert package_name.startswith(self.root_module_name)
            package_name = package_name[len(self.root_module_name) + 1:]
            logging.debug("Configuring parser for %s.%s", package_name, module_name)
            halp, desc = self.parse_help(module)
            data = {"type": "cmd", "help": halp, "desc": desc}
            if getattr(module, "__error__", False):
                data["error"] = True
            self.get_node(package_name)[module_name] = data

        if UNAVAILABLE_MODULES:
            logging.warning("The following modules are not available:")
            for name, msg in UNAVAILABLE_MODULES:
                logging.warning("    %s: %s", name, msg)

        command_map = {"module": self.root_module_name, "commands": self.root_node}

        if os.access(os.path.dirname(cached_command_map_filename), os.W_OK) and (
            not os.path.exists(cached_command_map_filename) or os.access(cached_command_map_filename, os.W_OK)
        ):
            logging.debug("Writing command map json to %s", cached_command_map_filename)

            cached_command_map_filename_new = cached_command_map_filename + ".new"
            with open(cached_command_map_filename_new, "w") as fpout:
                json.dump(command_map, fpout, indent=4)
            os.rename(cached_command_map_filename_new, cached_command_map_filename)
        else:
            logging.warning("Not updating commands cache (%s is not writeable)", cached_command_map_filename)
            logging.warning("Tip: are you using a shared install of octools? If so, no need to run oc refresh.")

        return command_map



#!/usr/bin/env python
# Copyright 2013 Netflix, Inc.

"""Utility classes
"""

from contextlib import contextmanager
import logging
import signal
import sys


class TimeoutError(Exception):
    """Timeout Error"""

    pass


@contextmanager
def timeout(seconds, error_message="Timeout"):
    """Timeout context manager using SIGALARM."""

    def _handle_timeout(signum, frame):  # pylint:disable=unused-argument,missing-docstring
        raise TimeoutError(error_message)

    if seconds > 0:
        signal.signal(signal.SIGALRM, _handle_timeout)
        signal.alarm(seconds)
    try:
        yield
    finally:
        if seconds > 0:
            signal.alarm(0)


def undecorate_func(func, decorators=None):
    """Finc the actual func behind any number of decorators
    """
    if decorators is None:
        decorators = []
    if hasattr(func, "original_func"):
        decorators.append(func)
        return undecorate_func(getattr(func, "original_func"), decorators)
    else:
        return func, decorators


def try_import(module_name):
    """Attempt to import the given module (by name), returning a tuple (True, module object) or (False,None) on ImportError"""
    try:
        module = __import__(module_name)
        return True, module
    except ImportError:
        return False, None


def call_chain(chain, *args, **kwargs):
    if len(chain) == 1:
        return chain[0](*args, **kwargs)
    elif len(chain) == 2:
        return chain[1](lambda: chain[0](*args, **kwargs))
    elif len(chain) == 3:
        return chain[2](lambda: chain[1](lambda: chain[0](*args, **kwargs)))
    else:
        raise Exception("call_chain is a hack and doesn't support chains longer than 3")


def profiling_wrapper(func):
    import cProfile
    import io
    import pstats

    pr = cProfile.Profile()
    pr.enable()
    try:
        func()
    finally:
        pr.disable()
        s = io.StringIO()
        sortby = "cumulative"
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())


def pdb_wrapper(func):
    try:
        return func()
    except Exception:
        import pdb
        import traceback

        etype, value, tb = sys.exc_info()
        logging.info("Top level exception caught, entering debugger")
        traceback.print_exc()
        pdb.post_mortem(tb)
        raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2013 Netflix, Inc.


"""Command line tools entry point."""

from cligraphy.core import capture, decorators, read_configuration, ctx
from cligraphy.core.log import setup_logging
from cligraphy.core.parsers import (
    AutoDiscoveryCommandMap,
    SmartCommandMapParser,
    ParserError,
    CustomDescriptionFormatter,
)
from cligraphy.core.reporting import ToolsPadReporter, NoopReporter
from cligraphy.core.util import undecorate_func, pdb_wrapper, profiling_wrapper, call_chain

from cachetools import cached

import faulthandler
try:
    from setproctitle import setproctitle  # pylint:disable=no-name-in-module
except ImportError:
    setproctitle = lambda x: None

import argcomplete
import argparse
import logging
import logging.config
import os
import signal
import subprocess
import sys


def _warn_about_bad_non_ascii_chars(args):
    """Detect non-ascii variants of some interesting characters, such as — instead of -"""
    bad_chars = ("—", "…", "“", "”", "\u200b")  # zero-width space
    try:
        line = " ".join(args)
        bad = [char in bad_chars for char in line]
        if any(bad):
            logging.warning(
                "Your command line contains %d bad unicode character(s)",
                len(bad),
            )
            logging.warning("> " + line)
            logging.warning("> " + "".join("^" if isbad else " " for isbad in bad))
            return
    except Exception:  # pylint:disable=broad-except
        logging.warning(
            "Exception while trying to detect bad unicode characters in command line args; continuing...", exc_info=True
        )


def _warn_about_bad_path(env_root, path):
    """Warn about mistakes in $PATH"""
    if not env_root:
        logging.warning("Running oc outside of its virtualenv is supported, but untested. Please report any bugs!")
        return
    env_bin = os.path.join(env_root, "bin")
    elements = path.split(":")
    try:
        position = elements.index(env_bin)
        if position != 0:
            logging.warning(
                "Your oc virtualenv, %s, is not the first element of your $PATH. Things might be broken.", env_root
            )
    except ValueError:
        logging.warning("Your oc virtualenv, %s, is not listed in your $PATH. Things are likely broken.", env_root)
        return


class _VersionAction(argparse.Action):
    """Shows last commit information"""

    def __call__(self, *args, **kwargs):
        try:
            os.chdir(ctx.conf.tool.repo_path)
            last_commit = subprocess.check_output(["git", "log", "-1"], stderr=subprocess.PIPE).decode('utf-8')
        except (subprocess.CalledProcessError, OSError):
            last_commit = "(could not get more recent commit information)"
        print("%s v%s\n%s" % (ctx.conf.tool.name, ctx.conf.tool.version, last_commit))
        sys.exit(1)


class Cligraph(object):
    reporter_cls = ToolsPadReporter

    def __init__(self, name, shortname, path):
        assert name
        assert shortname
        assert path
        self.tool_name = name
        self.tool_shortname = shortname
        self.tool_path = path
        self.conf, self.conf_layers = read_configuration(self)
        ctx.cligraph = self
        self.reporter = None

    def setup_reporter(self, args):
        """
        Setups up the reporter using ``self.reporter_cls`` class variable.
        NOTE: A reporter must be configured for Cligraphy to operate properly.
        If reporting is disabled via the config, a :class:`NoopReporter` is used instead.

        :param args: The parsed arguments for the command
        :type args: :class:`argparse.Namespace`
        :return: None
        """
        if self.conf.report.enabled:
            self.reporter = self.reporter_cls(self)
        else:
            self.reporter = NoopReporter()

    # FIXME(stf/oss): header names for octools
    def _setup_requests_audit_headers(self, command):
        """Setup requests user-agent and x-user/x-app headers.
        Just best effort - we don't care that much if this fails"""
        try:
            import requests

            def _default_user_agent(*args):
                return "requests (cligraphy/%s)" % command

            requests.utils.default_user_agent = _default_user_agent

            base_default_headers = requests.utils.default_headers

            def _default_headers(*args):
                headers = base_default_headers()
                headers["X-User"] = os.getenv("USER")
                headers["X-App"] = "cligraphy/%s" % command
                return headers

            requests.utils.default_headers = _default_headers
            requests.sessions.default_headers = _default_headers
        except Exception:
            logging.warn("Could set up requests audit headers, continuing anyway")
            pass

    # pylint:disable=protected-access
    def _run(self, args):
        """Run command by calling the main() function correctly (how we pass args depends on its actual signature).
        I the main method has been decorated, we need to look at the original function's argspec (but call the wrapper)
        """
        import inspect

        orig_func, _ = undecorate_func(args._func)
        argspec = inspect.getargspec(orig_func)
        if len(argspec.args) == 0:
            return args._func()
        elif argspec.varargs or argspec.keywords or len(argspec.args) > 1:
            kwargs = dict(**vars(args))
            for kw in list(kwargs.keys()):
                if kw.startswith("_"):
                    logging.debug("removing internal arg %s", kw)
                    del kwargs[kw]
            return args._func(**kwargs)
        else:
            if argspec.args[0] != "args":
                raise Exception(
                    'Programming error in command: if main() only has one argument it must be called "args"'
                )
            func = args._func
            for kw in list(vars(args).keys()):
                if kw.startswith("_") and kw not in (
                    "_parser",
                    "_cligraph",
                ):  # FIXME(stf/oss) maybe just expose cligraph
                    logging.debug("removing internal arg %s", kw)
                    delattr(args, kw)
            return func(args)

    def _run_command_process(self, args):
        """Command (child) process entry point. args contains the function to execute and all arguments."""

        setup_logging(args._level)

        command = " ".join(sys.argv[1:])
        setproctitle("oc/command/%s" % command)
        faulthandler.register(signal.SIGUSR2, all_threads=True, chain=False)  # pylint:disable=no-member
        self._setup_requests_audit_headers(command)

        ret = 1
        try:
            chain = [self._run]
            if args._profile:
                chain.append(profiling_wrapper)
            if args._pdb:
                chain.append(pdb_wrapper)
            ret = call_chain(chain, args)
        except SystemExit as exc:
            ret = exc.code
        except ParserError as pe:
            pe.report()
        except Exception:  # pylint:disable=broad-except
            logging.exception("Top level exception in command process")
        finally:
            sys.exit(ret)

    def _parse_args(self):
        """Setup parser and parse cli arguments.
        NB! counter-intuitively, this function also messes around with logging levels.
        """

        # We want some of our options to take effect as early as possible, as they affect command line parsing.
        # For these options we resort to some ugly, basic argv spotting

        if "--debug" in sys.argv:
            logging.getLogger().setLevel(logging.DEBUG)
            logging.debug("Early debug enabled")

        if "--verbose" in sys.argv or "-v" in sys.argv:
            logging.getLogger().setLevel(logging.INFO)

        autodiscover = False
        if "--autodiscover" in sys.argv:
            logging.debug("Autodiscover enabled")
            autodiscover = True

        parser = SmartCommandMapParser(
            prog=self.tool_shortname,
            description="Cligraphy command line tools",
            formatter_class=CustomDescriptionFormatter,
        )

        self.parser = parser  # expose to eg. ctx

        parser.add_argument("--version", action=_VersionAction, nargs=0, dest="_version")
        parser.add_argument(
            "--debug", help="enable debuging output", dest="_level", action="store_const", const=logging.DEBUG
        )
        parser.add_argument("--pdb", help="run pdb on exceptions", dest="_pdb", action="store_true")
        parser.add_argument(
            "--no-capture",
            help="(DEPRECATED) disable input/output capture",
            dest="_capture_deprecated",
            action="store_false",
            default=True,
        )  # DEPRECATED; left behind to avoid breaking existing references
        parser.add_argument(
            "--enable-capture", help="enable input/output capture", dest="_capture", action="store_true", default=False
        )
        parser.add_argument(
            "--no-reporting", help="disable reporting", dest="_reporting", action="store_false", default=True
        )
        parser.add_argument("--profile", help="enable profiling", dest="_profile", action="store_true", default=False)
        parser.add_argument(
            "--autodiscover",
            help="re-discover commands and refresh cache (default: read cached commands list)",
            dest="_autodiscover",
            action="store_true",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            help="enable informational output",
            dest="_level",
            action="store_const",
            const=logging.INFO,
        )

        for namespace, command_map in self.get_command_maps(autodiscover):
            parser.add_command_map(namespace, command_map)

        argcomplete.autocomplete(parser)

        _warn_about_bad_non_ascii_chars(sys.argv)
        _warn_about_bad_path(os.getenv("VIRTUAL_ENV"), os.getenv("PATH"))

        args = parser.parse_args()
        args._parser = parser  # deprecated

        # pylint:disable=protected-access
        if args._level is not None:
            logging.getLogger().setLevel(args._level)

        return args

    @cached(cache={})
    def get_command_maps(self, autodiscover=False):
        """Get all the command maps defined in our configuration.

        If autodiscover is True (defaults to False), python commands will be
        autodiscovered (instead of simply being obtained from a cached command map).

        :param autodiscover: (default - `False`) Whether or not to autodiscover commands
        :type autodiscover: bool
        :return: A dictionary containing the command map
        :rtype: dict
        """

        result = []
        for module, options in list(self.conf.commands.items()):
            try:
                if options is None:
                    options = {}
                logging.debug("Configuring commands module %s with options %s", module, options)

                opt_type = options.get("type", "python")
                opt_namespace = options.get("namespace", "")

                if opt_type == "python":
                    result.append(
                        (opt_namespace, AutoDiscoveryCommandMap(self, module).build(force_autodiscover=autodiscover))
                    )
                else:
                    raise Exception("Dont know how to handle commands module with type [%s]" % opt_type)
            except Exception as exc:  # pylint:disable=broad-except
                logging.warning(
                    "Could not configure commands module [%s] defined in configuration: %s. Skipping it.",
                    module,
                    exc,
                    exc_info=True,
                )

        return result

    def before_command_start(self, args, recorder):
        """Called after the args have been parsed but before the command starts
        Reports command start using the configured reporter, sets the process
        title to ``oc/parent/join script args``, and registers a fault handler

        :param args: The parsed args the command was called with
        :type args: :class:`argparse.Namespace`
        :param recorder: The output recorder being used to capture command output
        :type recorder: :class:`capture.Recorder`
        :rtype: None
        """
        # reporting: we send command executions report to a web service;
        #   unless the report.enabled conf key is false
        self.reporter.report_command_start(sys.argv)

        setproctitle("oc/parent/%s" % " ".join(sys.argv[1:]))
        faulthandler.register(signal.SIGUSR2, all_threads=True, chain=False)  # pylint:disable=no-member

    def after_command_finish(self, args, recorder, status):
        """Called after a command has finished running.
        Reports command exit and output using the configured reporter.

        :param args: The parsed args the command was called with
        :type args: :class:`argparse.Namespace`
        :param recorder: The output recorder being used to capture command output
        :type recorder: :class:`capture.Recorder`
        :param status: The exit status code of the command run
        :type status: int
        :rtype: None
        """
        # report execution details
        self.reporter.report_command_exit(status)
        self.reporter.report_command_output(recorder.output_as_string())  # TODO(stefan): also report total output size
        self.reporter.stop()

    def main(self):
        """Main oc wrapper entry point."""

        setup_logging()

        try:
            args = self._parse_args()
            logging.debug("Parsed args: %r", vars(args))
        except ParserError as pe:
            pe.report()

        if not os.isatty(sys.stdin.fileno()) or not os.isatty(sys.stdout.fileno()):
            logging.info("stdin or stdout is not a tty, disabling capture")
            args._capture = False

        self.setup_reporter(args)

        decs = decorators.get_tags(args._func)
        logging.debug("Decorator tags: %s", decs)

        if decorators.Tag.interactive in decs or not args._capture:
            recorder = capture.NoopOutputRecorder()
        else:
            recorder = capture.BufferingOutputRecorder(max_output_size=self.conf.report.max_output_size)

        self.before_command_start(args, recorder)

        if args._capture:
            # go ahead and run out command in a child process, recording all I/O
            logging.debug("Parent process %d ready to execute command process", os.getpid())
            status = capture.spawn_and_record(recorder, self._run_command_process, self.reporter.start, args)
            logging.debug("Command process exited with status %r", status)
        else:
            self.reporter.start()
            try:
                self._run_command_process(args)
            except SystemExit as exc:
                status = exc.code

        self.after_command_finish(args, recorder, status)

        return status



#!/usr/bin/env python
# Copyright 2015, 2016 Netflix, Inc.

from builtins import input


def prompt_int(prompt, value=None, default=None):
    """Prompt user for an int value"""
    while not isinstance(value, int):
        try:
            value = int(value)
        except TypeError:
            value = input(prompt)
            if value == "":
                value = default
    return value


def prompt_enter_choice(prompt, values, exceptions=None):
    prompt = "%s (%s)? " % (prompt, "/".join(values))
    values = {value.upper(): value for value in values}
    while True:
        try:
            value = input(prompt).upper()
        except BaseException as e:
            if exceptions and type(e) in exceptions:
                print()
                return exceptions[type(e)]
            else:
                raise
        if value in values:
            return values[value]



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

"""Unix terminal utils
"""

import fcntl
import termios
import struct
import sys
import os


def _ioctl_get_window_size(fd):
    """Calls TIOCGWINSZ for the given fd
    """
    try:
        return struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
    except IOError:
        return


def get_terminal_size():
    """Get current terminal size (best effort)
    """
    cr = _ioctl_get_window_size(0) or _ioctl_get_window_size(1) or _ioctl_get_window_size(2)
    if not cr:
        with os.open(os.ctermid(), os.O_RDONLY) as fd:
            cr = _ioctl_get_window_size(fd)
    return int(cr[0]), int(cr[1])


def set_terminal_size(lines, columns):
    """Set current terminal size
    """
    winsize = struct.pack("HHHH", lines, columns, 0, 0)
    fcntl.ioctl(1, termios.TIOCSWINSZ, winsize)
    sys.stdout.write("\x1b[8;{lines};{columns}t".format(lines=lines, columns=columns))



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

"""Session recording and replay
"""

from cligraphy.core import capture
from cligraphy.core.capture.termsize import get_terminal_size, set_terminal_size

import os.path
import os
import time


LOG_ROOT = "/tmp/session"


def session_capnp():
    """Load out capnproto schema
    """
    import capnp

    capnp.remove_import_hook()
    return capnp.load(os.path.join(capture.__path__[0], "session.capnp"))


class CapnpSessionRecorder(capture.Recorder):
    """Records a pty session
    """

    def __init__(self):
        self.last_ts = time.time()
        self.out_fp = None
        self.session_capnp = session_capnp()
        tm = time.gmtime()
        dirname = "%s/%04d/%02d/%02d/%02d" % (LOG_ROOT, tm.tm_year, tm.tm_mon, tm.tm_mday, tm.tm_hour)
        try:
            umask = os.umask(2)
            os.makedirs(dirname)
            os.umask(umask)
        except OSError:
            pass
        self.filename = dirname + "/%s-%d.log" % (os.getenv("USER"), time.time() * 1000)

    def start(self):
        """Start a session record
        """
        self.out_fp = open(self.filename, "wb")

        session = self.session_capnp.Session.new_message()

        session.username = os.getenv("USER")
        session.timestamp = int(time.time() * 1000)

        window_size = session.init("windowSize")
        window_size.lines, window_size.columns = get_terminal_size()

        environ = list(os.environ.items())
        session_env = session.init("environment", len(environ))
        for index, item in enumerate(environ):
            session_env[index].name = item[0]
            session_env[index].value = item[1]

        session.write(self.out_fp)
        self.out_fp.flush()
        self.last_ts = time.time()

    def _timecode(self):
        """Returns the current time code
        """
        now = time.time()
        ret = max(now - self.last_ts, 0.0)
        self.last_ts = now
        return ret

    def end(self, exitcode=0):
        """Finish a session record
        """
        event = self.session_capnp.Event.new_message()
        event.timecode = self._timecode()
        event.type = "sessionEnd"
        event.status = exitcode
        event.write_packed(self.out_fp)
        self.out_fp.close()

    def record_window_resize(self, lines, columns):
        """Record a window resizing event
        """
        event = self.session_capnp.Event.new_message()
        event.timecode = self._timecode()
        event.type = "windowResized"
        window_size = event.init("windowSize")
        window_size.lines, window_size.columns = lines, columns
        event.write_packed(self.out_fp)

    def record_user_input(self, data):
        """record user input separately from terminal output
        """
        event = self.session_capnp.Event.new_message()
        event.timecode = self._timecode()
        event.type = "userInput"
        event.data = data
        event.write_packed(self.out_fp)

    def record_server_output(self, data):
        """record terminal output (user + server originated)
        """
        event = self.session_capnp.Event.new_message()
        event.timecode = self._timecode()
        event.type = "ptyInput"
        event.data = data
        event.write_packed(self.out_fp)


class CapnpSessionPlayer(capture.Player):
    """Plays a recorded session
    """

    def __init__(self, filename):
        self.filename = filename
        self.session_capnp = session_capnp()
        self.fpin = open(self.filename, "rb")
        self.session = self.session_capnp.Session.read(self.fpin)

    def close(self):
        """Close this player
        """
        self.fpin.close()

    def play(self):
        """Generator of session playback events - basically (timecode, data) tuples
        """
        set_terminal_size(self.session.windowSize.lines, self.session.windowSize.columns)

        skipped = 0
        for event in self.session_capnp.Event.read_multiple_packed(self.fpin):
            if event.type == "ptyInput":
                yield event.timecode + skipped, event.data
                skipped = 0
            elif event.type == "windowResized":
                set_terminal_size(event.windowSize.lines, event.windowSize.columns)
                yield event.timecode + skipped, None
                skipped = 0
            else:
                skipped += event.timecode
        self.close()



#!/usr/bin/env python
# Copyright 2013, 2014 Netflix, Inc.

"""Command line tools entry point"""

from cligraphy.core.util import try_import

from cachetools import cached

import json
import logging
import logging.config
import os
import sys


#  new log level for, you guessed it, dryrun logging.
#  See: http://stackoverflow.com/a/13638084
DRYRUN_num = 15
DRYRUN_name = "DRYRUN"


def _dryrun(self, message, *args, **kws):
    if self.isEnabledFor(DRYRUN_num):
        self._log(DRYRUN_num, message, args, **kws)


class JsonHandler(logging.Handler):
    def __init__(self, destination, *args, **kwargs):
        self.destination = destination
        self.recurse = False
        super(JsonHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        if self.recurse:
            return
        try:
            self.recurse = True
            self._emit(record)
        finally:
            self.recurse = False

    def _emit(self, record, first_attempt=True):
        try:
            msg = record.msg % record.args
        except TypeError:
            msg = str(record.msg) + " (formatting error)"

        output = {}
        try:
            output = {
                "ts": int(record.created * 1000),
                "level": record.levelno,
                "msg": msg,
                # module, function, ...
                "module": record.module,
                "file": record.filename,
                "func": record.funcName,
                "line": record.lineno,
                # process and thread information
                "tid": record.thread,
                "tname": record.threadName,
                "pid": record.process,
                "pname": record.processName,
            }

            if hasattr(record, "extra"):
                if first_attempt:
                    output["extra"] = record.extra
                else:
                    output["extra"] = {
                        "_omitted": True
                    }  # We tried to serialize our record once and got an error, so omit ctx now

            if record.exc_info and record.exc_info[0] and record.exc_info[1]:
                output["exc_type"] = record.exc_info[0].__name__
                output["exc_msg"] = repr(record.exc_info[1].message)
                tb = []
                cur = record.exc_info[2]
                while cur:
                    frame = cur.tb_frame
                    tb.append((frame.f_code.co_filename, frame.f_code.co_name, frame.f_lineno))
                    cur = cur.tb_next
                output["exc_tb"] = tb

            try:
                msg = json.dumps(output)
                self.destination.log_event("log", record.created, msg)
            except (TypeError, OverflowError):
                if first_attempt:
                    return self.emit(record, False)  # Try again, without ctx this time
                else:
                    raise

        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            print("log.structured: serialization error - details follow")
            print(output)
            self.handleError(record)


def silence_verbose_loggers():
    logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.WARN)


@cached(cache={})
def _get_dict_config():
    dict_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {
                "datefmt": "%Y-%m-%d %H:%M:%S",
                "format": "%(asctime)s %(levelname)-8s %(module)s %(funcName)s %(message)s",
            }
        },
        "handlers": {"defaultHandler": {"level": "DEBUG", "class": "logging.StreamHandler", "formatter": "default"}},
        "loggers": {"": {"handlers": (["defaultHandler"]), "level": "WARN"}},
    }

    if os.isatty(sys.stderr.fileno()) and try_import("colorlog")[0]:
        import colorlog

        colorlog.default_log_colors.update({DRYRUN_name: "blue"})
        dict_config["formatters"]["colors"] = {
            "()": "colorlog.ColoredFormatter",
            "datefmt": "%Y-%m-%d %H:%M:%S",
            "format": "%(log_color)s%(asctime)s %(levelname)-8s%(reset)s %(purple)s%(module)s/%(funcName)s%(reset)s %(message)s",
        }
        dict_config["handlers"]["defaultHandler"]["formatter"] = "colors"

    return dict_config


def setup_logging(level=None):
    """Configure logging"""
    logging.addLevelName(DRYRUN_num, DRYRUN_name)
    logging.Logger.dryrun = _dryrun
    try:
        logging.config.dictConfig(_get_dict_config())
        logging.captureWarnings(True)
        silence_verbose_loggers()
        if level is not None:
            logging.getLogger().setLevel(level)
    except Exception:  # pylint:disable=broad-except
        logging.basicConfig(level=logging.WARN)
        logging.warn("Could not configure logging, using basicConfig", exc_info=True)



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

"""Terminal session capture"""

from cligraphy.core.capture import ptysnoop

from abc import abstractmethod


class Recorder(object):
    def start(self):
        pass

    def end(self, exitcode=0):
        """Finish a session record
        """

    def record_window_resize(self, lines, columns):
        """Record a window resizing event
        """

    @abstractmethod
    def record_user_input(self, data):
        """record user input separately from terminal output
        """

    @abstractmethod
    def record_server_output(self, data):
        """record terminal output (user + server originated)
        """


class Player(object):
    """Plays a recorded session
    """

    @abstractmethod
    def close(self):
        """Close this player
        """

    @abstractmethod
    def play(self):
        """Generator of session playback events - basically (timecode, data) tuples
        """


class NoopOutputRecorder(Recorder):
    def output_as_string(self):
        return "(output discarded)"


class BufferingOutputRecorder(Recorder):
    """Recorder that buffers output up to a maximum size"""

    def __init__(self, max_output_size):
        super(BufferingOutputRecorder, self).__init__()
        self._remaining_size = max_output_size
        self._total_size = 0
        self._buffer = []

    @abstractmethod
    def record_server_output(self, data):
        """record terminal output (user + server originated)
        """
        self._total_size += len(data)
        if self._remaining_size > 0:
            part = data[: self._remaining_size]
            self._buffer.append(part)
            self._remaining_size = max(0, self._remaining_size - len(part))

    @property
    def total_size(self):
        return self._total_size

    def output_as_string(self):
        return "".join(self._buffer)


def spawn_and_record(recorder, func, parent_func, *args, **kwargs):
    script = ptysnoop.Script(recorder)
    return script.run(func, parent_func, *args, **kwargs)



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

"""
Cligraphy tools
"""

from attrdict import AttrDict

import yaml

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

import collections
import os
import os.path
import re
import logging


class Context(object):
    def __init__(self):
        self._cligraph = None

    @property
    def cligraph(self):
        return self._cligraph

    @cligraph.setter
    def cligraph(self, value):
        self._cligraph = value

    @property
    def conf(self):
        return self._cligraph.conf

    @property
    def parser(self):
        return self._cligraph.parser


ctx = Context()


__CONF_SUBST_RE = re.compile(r"(%cfg.([a-zA-Z0-9.-_]+)%)")


def dictify_recursive(obj):
    """Transforms an attrdict tree into regular python dicts"""
    for key, val in obj.items():
        if isinstance(val, AttrDict):
            obj[key] = dictify_recursive(val)
    return dict(**obj)


def update_recursive(base, overlay):
    """Resursively update base with values from overlay
    """
    for key, val in overlay.items():
        if isinstance(val, collections.Mapping):
            base[key] = update_recursive(base.get(key, {}), val)
        else:
            base[key] = overlay[key]
    return base


def find_node(node, path, add=False):
    """Find a node in a dict tree. Path is the broken down path (array). Add missing nodes if add=True."""
    for part in path:
        if not node or part not in node:
            if add:
                node[part] = {}
            else:
                return None
        node = node[part]
    return node


def get(root, confkey):
    """Config key getter"""
    return find_node(root, confkey.split("."))


def _resolve_config(root, subst_re, node=None):
    """recursive helper for resolve_config"""
    if node is None:
        node = root
    remaining = []
    for key, val in node.items():
        if isinstance(val, collections.Mapping):
            remaining.extend(_resolve_config(root, subst_re, node=val))
        elif isinstance(val, str):
            for match, confkey in subst_re.findall(val):
                confval = get(root, confkey)
                if confval is not None and (not isinstance(confval, str) or subst_re.search(confval) is None):
                    node[key] = val = val.replace(match, str(confval))
                else:
                    remaining.append("%s: %s" % (key, val))
    return remaining


def resolve_config(cfg):
    """performs variable substitution in a configuration tree"""
    remaining = []
    for _ in range(16):  # 16 maximum substitution passes
        remaining_prev = remaining
        remaining = _resolve_config(cfg, __CONF_SUBST_RE)
        if not remaining or len(remaining) == len(remaining_prev):
            break
    if remaining:
        raise Exception("Incorrect configuration file: could not resolve some configuration variables:  %s" % remaining)


def automatic_configuration(cligraph, layer_name):
    auto_data = {
        "tool": {
            "name": cligraph.tool_name,
            "shortname": cligraph.tool_shortname,
            "version": "1.0",
            "repo_path": cligraph.tool_path,
        },
        "repos": {"root": os.path.abspath(os.path.join(cligraph.tool_path, ".."))},
    }
    username = os.getenv("USER")
    if username:
        auto_data["user"] = {"name": username, "email": "%s@domain.net" % (username)}  # OPEN SOURCE TODO
    else:
        auto_data["user"] = {"name": "unknown", "email": "octools-unknown-user@domain.net"}  # OPEN SOURCE TODO
    auto_data["user"]["dotdir"] = os.path.abspath(os.path.expanduser("~/." + cligraph.tool_shortname))
    return auto_data


def read_configuration(cligraph, custom_suffix=""):
    """Read configuration dict for the given tool
    """

    cfg = {}
    layers = collections.OrderedDict()
    layers["auto"] = [automatic_configuration, None]
    layers["shared"] = [os.path.join(cligraph.tool_path, "conf/%s.yaml" % cligraph.tool_shortname), None]
    layers["custom"] = [
        os.path.join(
            os.path.abspath(os.path.expanduser("~/." + cligraph.tool_shortname)),
            "%s.yaml%s" % (cligraph.tool_shortname, custom_suffix),
        ),
        None,
    ]

    for layer_name, layer_data in list(layers.items()):
        if callable(layer_data[0]):
            layer = layer_data[0](cligraph, layer_name)
        else:
            if not os.path.exists(layer_data[0]):
                continue
            with open(layer_data[0], "r") as filep:
                layer = yaml.load(filep, Loader=Loader)
        layers[layer_name][1] = layer
        if layer:
            update_recursive(cfg, layer)

    resolve_config(cfg)
    return AttrDict(**cfg), layers


def write_configuration_file(filename, conf):
    """Write a config dict to the specified file, in yaml format"""
    import shutil

    try:
        shutil.copy2(filename, "%s.back" % filename)
    except IOError:
        pass

    try:
        os.makedirs(os.path.dirname(filename))
    except OSError:
        pass

    with open(filename, "w") as filep:
        yaml.dump(conf, filep, indent=4, default_flow_style=False, Dumper=Dumper)


def edit_configuration(tool_name, callback):
    """Wrapper for configuration editing tools
    """
    import shutil

    filename = os.path.join(USER_DOTDIR, "%s.yaml" % tool_name)
    edit_filename = "%s.edit" % filename

    if os.path.exists(edit_filename):
        logging.warn("Unfinished edit file %s exists, editing that one.", edit_filename)
    else:
        if os.path.exists(filename):
            shutil.copy2(filename, edit_filename)
        else:
            with open(edit_filename, "w") as fp:
                fp.write("\n")

    callback(edit_filename)

    read_configuration(tool_name, custom_suffix=".edit")  # validate that configuration is still readable

    if os.path.exists(filename):
        shutil.copy2(filename, "%s.back" % filename)
    os.rename(edit_filename, filename)



#!/usr/bin/env python
# Copyright 2013 Netflix, Inc.

"""Decorators."""

import cligraphy.core.parsers
from cligraphy.core.util import undecorate_func

from builtins import input
from functools import wraps
import logging

from enum import Enum


class Tag(Enum):
    """Enumeration of the decorator tags we support."""

    beta = 1
    disabled = 2
    strict = 3
    interactive = 4


class DisabledCommandException(Exception):
    """Raised when a @disabled command is invoked."""

    pass


def tag(tag_enum, reason="No reason given"):
    """Decorator that disables a command."""
    # pylint:disable=missing-docstring,unused-argument
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            func(*args, **kwargs)

        wrapper.tag = tag_enum
        wrapper.original_func = func
        return wrapper

    return actual_decorator


def disabled(reason="No reason given"):
    """Decorator that disables a command."""
    # pylint:disable=missing-docstring,unused-argument
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            raise DisabledCommandException("This command is disabled: %s" % reason)

        wrapper.tag = Tag.disabled
        wrapper.original_func = func
        return wrapper

    return actual_decorator


def beta(reason="No reason given"):
    """Decorator for unstable commands - user will be prompted to confirm before execution."""
    # pylint:disable=missing-docstring,unused-argument
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            message = (
                "WARNING - this command is marked as beta: %s - are you sure you want to continue (y/N)? " % reason
            )
            logging.warning(message)
            confirm = input(message)
            if confirm and confirm.lower() in ("y", "yes"):
                func(*args, **kwargs)
            else:
                print("Cancelled")

        wrapper.tag = Tag.beta
        wrapper.original_func = func
        return wrapper

    return actual_decorator


def strict(reason="No reason given"):
    """Command decorator that disables fuzzy matching."""
    # pylint:disable=missing-docstring,unused-argument
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if cligraphy.core.parsers.FUZZY_PARSED:
                logging.error(
                    "This command must be called by its exact name (%s). Please run as %s",
                    reason,
                    cligraphy.core.parsers.FUZZY_PARSED[0][-1],
                )
                return
            else:
                func(*args, **kwargs)

        wrapper.tag = Tag.strict
        wrapper.original_func = func
        return wrapper

    return actual_decorator


def get_tags(decorated_func):
    """Return the set of decorator tags applied to decorated_func."""
    _, decorators = undecorate_func(decorated_func)
    return set([x.tag for x in decorators])



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

"""Human-friendly time period parsing
"""

import time
import math
from datetime import datetime
from dateutil import parser

UNITS = {
    "month": 3600 * 24 * 31,  # FIXME 1 month is not really 31 days
    "week": 3600 * 24 * 7,
    "day": 3600 * 24,
    "hour": 3600,
    "minute": 60,
    "second": 1,
}

UNITS["mon"] = UNITS["month"]
UNITS["w"] = UNITS["week"]
UNITS["d"] = UNITS["day"]
UNITS["h"] = UNITS["hour"]
UNITS["min"] = UNITS["minute"]
UNITS["m"] = UNITS["minute"]
UNITS["sec"] = UNITS["second"]
UNITS["s"] = UNITS["second"]


def parse_time(human_time):
    """Parse a human-friendly time expression, such as 1d or now, and return a unix timestamp
    """
    human_time = human_time.lower()
    now = int(time.time())
    if human_time == "now":
        return now
    try:
        unit = human_time[-1]
        value = int(human_time[:-1])
        return now - (value * UNITS[unit])
    except Exception:
        raise Exception("Dont know what to do with time [%s]" % human_time)


def parse_datetime(human_time):
    """Parse a human-friendly time expression (such as 1d or now)
            or datetime formatted as string (such as 2016-01-20 20:00)
        and return a datetime object
    """
    try:
        return datetime.fromtimestamp(parse_time(human_time))
    except Exception:
        try:
            dt = parser.parse(human_time)
            return dt
        except Exception:
            raise Exception("Dont know what to do with time [%s]" % human_time)


def align_times(start, end, ratio=0.02):
    """Align timestamps on a reasonable boundary
    """
    assert start < end, "Start time must be before end time"
    boundary = int(math.ceil((end - start) * ratio))
    return int(start / boundary) * boundary, int(end / (boundary + 1)) * (boundary + 1)



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

from cligraphy.core.tracking import TRACKING

import threading
from queue import Queue, Empty, Full
import collections

import requests

import logging
import os
import socket
import time

ReportingEvent = collections.namedtuple("ReportingEvent", "name,args,kwargs")


class Reporter(object):
    reporting_event_cls = ReportingEvent

    def report_command_start(self, command_line):
        """Report command execution"""
        pass

    def report_command_output(self, output):
        """Report command output"""
        pass

    def report_command_exit(self, exit_code):
        """Report command exit code"""
        pass

    def start(self):
        """Start this reporter"""
        pass

    def stop(self):
        """Stop this reporter"""
        pass


class NoopReporter(Reporter):
    """Reporter that does nothing"""


class ThreadedReporter(Reporter, threading.Thread):
    """Base class for reporters"""

    idle_period = 10

    def __init__(self):
        super(Reporter, self).__init__(group=None, name="oc-reporter-thread")
        self.daemon = True
        self.queue = Queue(maxsize=16)
        self.stop_event = threading.Event()

    def run(self):
        logging.debug("Starting reporter thread mainloop")
        while True:
            try:
                event = self.queue.get(block=True, timeout=self.idle_period)
                if event:
                    if event.name == "stop":
                        return
                    func = getattr(self, "_report_%s" % event.name, None)
                    if func:
                        func(*event.args, **event.kwargs)
            except Empty:
                if self.stop_event.is_set():
                    return
                else:
                    try:
                        self._report_idle()
                    except Exception:
                        pass

    def _report(self, event_name, *args, **kwargs):
        event = self.reporting_event_cls(name=event_name, args=args, kwargs=kwargs)
        try:
            self.queue.put(event, block=False)
        except Full:
            logging.debug("Could not put event (name=%s) in reporting queue", event_name, exc_info=True)

    def _report_idle(self):
        pass

    def report_command_start(self, command_line):
        """Report command execution"""
        self._report("command_start", command_line)

    def report_command_output(self, output):
        """Report command output"""
        self._report("command_output", output)

    def report_command_exit(self, exit_code):
        """Report command exit code"""
        self._report("command_exit", exit_code)

    def start(self):
        threading.Thread.start(self)

    def stop(self, timeout=1.5):
        self.stop_event.set()
        self._report("stop")
        logging.debug("Waiting on reporter thread for up to %s second", timeout)
        self.join(timeout=timeout)
        if self.isAlive():
            logging.debug("Reported thread is still running")


def disable_ssl_warnings():
    import requests.packages.urllib3

    requests.packages.urllib3.disable_warnings()


class ToolsPadReporter(ThreadedReporter):

    timeout = 0.75

    def __init__(self, cligraph):
        super(ToolsPadReporter, self).__init__()
        self.cligraph = cligraph
        self.requests_session = requests.Session()
        self.requests_session.headers.update(
            {"User-Agent": "octools:%s@%s" % (os.getenv("USER"), socket.gethostname())}
        )
        self.created = False
        self.session_endpoint = None
        self.create_event_endpoint = None
        self.idle_endpoint = "%s" % self.cligraph.conf.report.server
        self.create_session_endpoint = "%s/api/v0.1/execution" % self.cligraph.conf.report.server
        disable_ssl_warnings()

    def _report_idle(self):
        self.requests_session.head(self.idle_endpoint)

    def _report_command_start(self, command_line):
        try:
            response = self.requests_session.post(
                self.create_session_endpoint,
                data={
                    "uuid": TRACKING.execution_uuid,
                    "session_uuid": TRACKING.session_uuid,
                    "user_email": self.cligraph.conf.user.email,
                    "start_stamp": int(time.time()),
                    "command_line": " ".join(command_line),
                },
                timeout=self.timeout,
            )
            if response.status_code == 201:
                self.created = True
                self.session_endpoint = "%s%s" % (self.cligraph.conf.report.server, response.json()["uri"])
                self.create_event_endpoint = "%s/event" % (self.session_endpoint)
            else:
                response.raise_for_status()

        except requests.exceptions.RequestException:
            logging.debug("Could not report command start", exc_info=True)

    def _report_command_output(self, output):
        if not self.created:
            return
        try:
            response = self.requests_session.post(
                self.create_event_endpoint,
                data={"stamp": int(time.time()), "type": "output", "body": output},
                timeout=self.timeout,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException:
            logging.debug("Could not report comand output", exc_info=True)

    def _report_command_exit(self, exit_code):
        if not self.created:
            return
        try:
            response = self.requests_session.patch(
                self.session_endpoint,
                data={"end_stamp": int(time.time()), "exit_code": exit_code},
                timeout=self.timeout,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException:
            logging.debug("Could not report command exit", exc_info=True)



#!/usr/bin/env python
# Copyright 2014 Netflix, Inc.

"""Capture interactive unix terminal activity

Basically a python reimplementation of ye old 'script' utility
"""

import errno
import fcntl
import logging
import os
import select
import signal
import struct
import termios
import threading
import multiprocessing
import time

STDIN, STDOUT, STDERR = 0, 1, 2
DEFAULT_BUFF_SIZE = 128
WAKEUP = "!"


def xwrite(fileno, data):
    """Write data to fileno
    """
    offset = 0
    remaining = len(data)
    while remaining > 0:
        count = os.write(fileno, data[offset:])
        remaining -= count
        offset += count


class Script(object):
    """A python reimplementation of the script utility
    """

    def __init__(self, recorder, buff_size=DEFAULT_BUFF_SIZE, idle_timeout=0, select_timeout=1):
        self.recorder = recorder
        self.buff_size = buff_size
        self.idle_timeout = idle_timeout
        self.select_timeout = select_timeout
        self.activity_stamp = time.time()
        self.master = None
        self.slave = None
        self.child_pid = None
        self.start_event = multiprocessing.Event()
        self.stop_event = threading.Event()
        self.resize_event = threading.Event()
        self.tcattr = None
        self.wakeup_r, self.wakeup_w = os.pipe()

    def _on_stdin_input(self):
        """User typing something
        """
        data = os.read(STDIN, self.buff_size)

        if len(data) == 0:  # EOF
            self.stop_event.set()
        else:
            xwrite(self.master, data)
            self.recorder.record_user_input(data)

    def _on_pty_input(self):
        """Some data has been displayed on screen
        """
        data = os.read(self.master, self.buff_size)

        if len(data) == 0:
            self.stop_event.set()
        else:
            xwrite(1, data)
            self.recorder.record_server_output(data)

    def _on_wakeup(self):
        """Our main thread woke us up
        """
        os.read(self.wakeup_r, 1)

        if self.resize_event.is_set():
            self.resize_event.clear()
            data = fcntl.ioctl(STDIN, termios.TIOCGWINSZ, "0123")
            lines, columns = struct.unpack("hh", data)
            self.recorder.record_window_resize(lines, columns)
            fcntl.ioctl(self.slave, termios.TIOCSWINSZ, data)

    def _io_loop(self):
        """I/O loop (executed as a parent process thread)
        """

        self.start_event.set()

        io_actions = {STDIN: self._on_stdin_input, self.master: self._on_pty_input, self.wakeup_r: self._on_wakeup}

        rlist = list(io_actions.keys())

        self.recorder.start()

        while not self.stop_event.is_set():
            try:
                activity = select.select(rlist, [], [], self.select_timeout)[0]
            except select.error as err:
                assert err.args[0] != errno.EINTR, "Should not be getting interrupted syscalls in thread"
                raise

            if activity:
                self.activity_stamp = time.time()
            elif self.idle_timeout > 0:
                if time.time() > self.activity_stamp + self.idle_timeout:
                    self.stop_event.set()

            for active_fd in activity:
                try:
                    io_actions[active_fd]()
                except OSError as ose:
                    assert ose.errno != errno.EINTR, "Should not be getting interrupted syscalls in thread"
                    raise

    def _on_sigchild(self, *args, **kwargs):  # pylint:disable=unused-argument
        """SIGCHILD handler
        """
        signal.signal(signal.SIGCHLD, self._on_sigchild)
        # logging.debug('Got sigchild')
        self.stop_event.set()
        os.write(self.wakeup_w, WAKEUP)

    def _on_sigwinch(self, *args, **kwargs):  # pylint:disable=unused-argument
        """SIGWINCH handler
        """
        # logging.debug('Got sigwinch')
        self.resize_event.set()
        os.write(self.wakeup_w, WAKEUP)

    def _run_parent(self):
        """Parent process main loop
        """
        io_thread = threading.Thread(group=None, target=self._io_loop, name="ptysnoop_io_thread")
        io_thread.start()

        signal.signal(signal.SIGCHLD, self._on_sigchild)
        signal.signal(signal.SIGWINCH, self._on_sigwinch)

        while not self.stop_event.is_set():
            self.stop_event.wait(10)

        io_thread.join()

    def _open_pty(self):
        """Create a PTY
        """
        # get our terminal params
        self.tcattr = termios.tcgetattr(STDIN)
        winsize = fcntl.ioctl(STDIN, termios.TIOCGWINSZ, "0123")
        # open a pty
        self.master, self.slave = os.openpty()
        # set the slave's terminal params
        termios.tcsetattr(self.slave, termios.TCSANOW, self.tcattr)
        fcntl.ioctl(self.slave, termios.TIOCSWINSZ, winsize)

    def _setup_slave_pty(self):
        """Set suitable tty options for our pty slave
        """
        os.setsid()
        fcntl.ioctl(self.slave, termios.TIOCSCTTY, 0)
        os.close(self.master)
        os.dup2(self.slave, STDIN)
        os.dup2(self.slave, STDOUT)
        os.dup2(self.slave, STDERR)  # FIXME can we handle stderr better?
        os.close(self.slave)

    def _fix_tty(self):
        """Set suitable tty options
        """
        assert self.tcattr is not None
        iflag, oflag, cflag, lflag, ispeed, ospeed, chars = self.tcattr  # pylint:disable=unpacking-non-sequence
        # equivalent to cfmakeraw
        iflag &= ~(
            termios.IGNBRK
            | termios.BRKINT
            | termios.PARMRK
            | termios.ISTRIP
            | termios.INLCR
            | termios.IGNCR
            | termios.ICRNL
            | termios.IXON
        )
        oflag &= ~termios.OPOST
        lflag &= ~(termios.ECHO | termios.ECHONL | termios.ICANON | termios.ISIG | termios.IEXTEN)
        cflag &= ~(termios.CSIZE | termios.PARENB)
        cflag |= termios.CS8
        termios.tcsetattr(STDIN, termios.TCSANOW, [iflag, oflag, cflag, lflag, ispeed, ospeed, chars])

    def _done(self):
        """Nicely close our logs, reset the terminal and exit
        """
        os.close(self.master)
        termios.tcsetattr(STDIN, termios.TCSADRAIN, self.tcattr)

        pid, status = os.waitpid(self.child_pid, os.WNOHANG)
        if pid != 0:
            assert pid == self.child_pid
            exit_code = status >> 8
            self.recorder.end(exit_code)
            return exit_code
        else:
            logging.warn("waitpid(%d) returned %d %d", self.child_pid, pid, status)

    def run(self, callback, parent_callback, *args, **kwargs):
        """Setup, fork and run
        """

        if threading.active_count() > 1:
            threads = threading.enumerate()
            logging.warning("Programming error: there are %d active threads (list follows)", len(threads))
            for thread in threads:
                logging.warning("  - %s", thread)
            logging.warning("Creating threads before forking can lead to issues and should be avoided")

        self._open_pty()
        self._fix_tty()

        pid = os.fork()
        if pid == 0:
            # child
            self._setup_slave_pty()
            self.start_event.wait()
            callback(*args, **kwargs)
        else:
            # parent
            try:
                self.child_pid = pid
                if parent_callback:
                    parent_callback()
                self._run_parent()
            finally:
                return self._done()



#!/usr/bin/env python
# Copyright 2013-2018 Netflix, Inc.

from setuptools import setup

setup(
    name = 'cligraphy',
    version = '0.1.0',
    description = 'Cligraphy Command line tools',
    long_description = 'Cligraphy Command line tools',
    author = 'Netflix, Inc.',
    author_email = '',  # OPEN SOURCE TODO
    include_package_data=True,
    license='Apache 2.0',
    zip_safe=False,
    setup_requires=[
        'setupmeta'
    ],
    extras_require={
        'dev': [
            'black',
            'check-manifest',
            'flake8',
            'pytest',
            'pycodestyle',
        ],
    },
)
